use crate::{fleck_core::MessageParams, prelude::*};

use serde::{Deserialize, Serialize};
use std::{cell::RefCell, collections::HashSet, rc::Rc};

#[derive(Serialize, Deserialize, Debug)]
struct DiscoveryMsg {
    pkey: fleck_core::crypto::PublicKey,
    ttl: u8,
}

impl fleck_core::msg::MessageParams for DiscoveryMsg {
    const NAME: &'static str = "DiscoveryMsg";
    const ENCRYPTED: bool = false;
    const SIGNED: bool = false;
}

/// Handles incoming DiscoveryMsgs
pub(crate) struct DiscoveryService {
    api: Rc<crate::API>,
}

impl Service for DiscoveryService {
    fn new(api: Rc<crate::API>) -> Self {
        Self { api }
    }

    fn setup(self: &Rc<Self>) {
        self.api.with_service(|msg: &fleck_core::MessageService| {
            msg.add_message_type::<DiscoveryMsg>();
        });

        self.api
            .channel::<fleck_core::msg::MessageChannel<DiscoveryMsg>>()
            .sub_eat(self, Self::discovery);
    }
}

impl DiscoveryService {
    fn discovery(&self, msg: (fleck_core::msg::Metadata, DiscoveryMsg)) {
        let discovery = msg.1;

        // this will automatically ignore self-messages because we already know our own pubkey
        self.api.with_service(|ns: &fleck_core::NodeService| {
            ns.inform_of(discovery.pkey, msg.0.peer);
        });
    }
}

/// Periodically sends out multicast messages to look for peers on local network(s)
pub struct LocalDiscovery {
    api: Rc<crate::API>,
    targets: RefCell<Vec<fleck_core::peer::Peer>>,
}

impl Service for LocalDiscovery {
    fn new(api: Rc<crate::API>) -> Self {
        Self {
            api,
            targets: Default::default(),
        }
    }
    fn setup(self: &Rc<Self>) {
        self.api
            .channel::<fleck_core::MajorTickChannel>()
            .sub_ref(self, LocalDiscovery::process_major_tick);
    }
}

impl LocalDiscovery {
    pub fn add_multicast_peer(&self, peer: fleck_core::peer::Peer) {
        // TODO: verify that the peer is in fact a multicast address...
        self.targets.borrow_mut().push(peer);
    }

    fn process_major_tick(&self, _: &mut ()) {
        for peer in self.targets.borrow().iter() {
            let msg = DiscoveryMsg {
                pkey: self.api.with_service(|ns: &fleck_core::NodeService| {
                    ns.self_node().pubkey().to_owned()
                }),
                ttl: 1,
            }
            .into_message()
            .with_peer(peer.clone());

            self.api.queue::<fleck_core::SendPacketChannel>(msg);
        }
    }
}

/// Queries peers that have sent us a message for their pubkeys if not already known;
/// can also be used to introduce a remote peer by address.
pub struct PeerDiscovery {
    api: Rc<crate::API>,
    known: RefCell<HashSet<fleck_core::peer::Peer>>,
}

impl Service for PeerDiscovery {
    fn new(api: Rc<crate::API>) -> Self {
        Self {
            api,
            known: Default::default(),
        }
    }

    fn setup(self: &Rc<Self>) {
        self.api
            .channel::<fleck_core::ReceivePacketChannel>()
            .sub_opt(fleck_core::ReceiveOrder::Annotate, self, Self::discover);
    }
}

impl PeerDiscovery {
    fn discover(&self, msg: fleck_core::Message) -> Option<fleck_core::Message> {
        let peer = match &msg.peer {
            Some(peer) => peer,
            // don't touch if it doesn't have a peer
            None => return Some(msg),
        };
        // check to see if we know this peer yet
        if self.known.borrow().contains(peer) {
            self.new_peer(peer.clone());
        }

        Some(msg)
    }

    /// Introduce a new peer into the `fleck` network.
    ///
    /// This function takes a [`fleck_core::Peer`] instance, which may be difficult to get ahold of
    /// in some circumstances. See [`fleck_core::io`] for details, but generally, you can retrieve
    /// a `Peer` from whatever [`fleck_core::io::FdHandler`] is providing communication.
    pub fn new_peer(&self, peer: fleck_core::Peer) {
        let mut known = self.known.borrow_mut();
        known.insert(peer.clone());

        log::trace!("given new peer!");

        self.api.queue::<fleck_core::SendPacketChannel>(
            DiscoveryMsg {
                pkey: self
                    .api
                    .with_service(|ns: &fleck_core::NodeService| ns.self_node().pubkey().clone()),
                ttl: 2,
            }
            .into_message()
            .with_peer(peer),
        );
    }
}
