use super::{fleck_core, FdHandler, IOService, InterestRegistration};
use std::{os::unix::prelude::AsRawFd, rc::Rc};

/// Build a `timerfd` instance to receive regular updates in a [`crate::Service`]
#[derive(Default)]
pub struct TimerFdBuilder {
    interval: Option<std::time::Duration>,
    trigger: Option<Box<dyn Fn()>>,
}

impl TimerFdBuilder {
    pub fn interval(mut self, interval: std::time::Duration) -> Self {
        self.interval = Some(interval);
        self
    }

    pub fn trigger<F: 'static + Fn()>(mut self, f: F) -> Self {
        self.trigger = Some(Box::new(f));
        self
    }

    pub fn build(self, io: &IOService) {
        let mut timer = timerfd::TimerFd::new().expect("couldn't create new timerfd?");

        timer.set_state(
            timerfd::TimerState::Periodic {
                current: std::time::Duration::from_millis(1),
                interval: self
                    .interval
                    .expect("building timerfd with no interval set?"),
            },
            timerfd::SetTimeFlags::Default,
        );

        let interest = io.register_interest(timer.as_raw_fd());

        let sock = Rc::new(TimerFd {
            interest,
            timer,
            trigger: self.trigger.expect("building timerfd with no trigger?"),
        });

        io.api.queue::<super::NewHandlerChannel>(sock);
    }
}

struct TimerFd {
    interest: InterestRegistration,
    timer: timerfd::TimerFd,
    trigger: Box<dyn Fn()>,
}

impl FdHandler for TimerFd {
    fn interest(&self) -> &InterestRegistration {
        &self.interest
    }
    fn ready_read(&self, _rc: &Rc<dyn FdHandler>) {
        self.timer.read();
        (self.trigger)();
    }

    fn ready_write(&self, _rc: &Rc<dyn FdHandler>) {
        unreachable!()
    }

    fn dispatch(&self, _msg: fleck_core::msg::Message) {
        unreachable!()
    }
}
