use rand::prelude::*;

use std::{
    cell::RefCell,
    collections::HashMap,
    rc::{Rc, Weak},
};

use crate::helper::InnerRef;
use crate::prelude::*;

use fleck_core::crypto::{Keypair, PublicKey};

#[derive(Debug, Default)]
pub struct Node {
    pubkey: PublicKey,
    peer: Option<fleck_core::peer::Peer>,
    keypair: RefCell<Option<Keypair>>,
}

impl Node {
    pub fn build(pubkey: PublicKey, peer: Option<fleck_core::peer::Peer>) -> Self {
        Self {
            pubkey,
            peer,
            ..Default::default()
        }
    }

    pub fn build_ephemeral() -> Self {
        let mut csprng = StdRng::from_entropy();
        let kp = Keypair::generate(&mut csprng);

        Self {
            pubkey: PublicKey::wrap(kp.public),
            peer: None,
            keypair: RefCell::new(Some(kp)),
        }
    }

    pub fn build_with_keypair(kp: Keypair) -> Self {
        Self {
            pubkey: PublicKey::wrap(kp.public),
            peer: None,
            keypair: RefCell::new(Some(kp)),
        }
    }

    pub fn peer(&self) -> Option<fleck_core::peer::Peer> {
        self.peer.clone()
    }

    pub fn pubkey(&self) -> &PublicKey {
        &self.pubkey
    }
    pub(crate) fn keypair(&self) -> Option<InnerRef<'_, Keypair>> {
        InnerRef::build(self.keypair.borrow())
    }
}

pub struct RegTag {}
pub type NodeRegistrationChannel = (RegTag, Rc<Node>);

type NodePeerMap = HashMap<fleck_core::Peer, Rc<Node>>;
pub struct NodeService {
    api: Rc<crate::API>,
    node_by_pubkey: RefCell<HashMap<fleck_core::crypto::PublicKey, Rc<Node>>>,
    node_by_peer: RefCell<NodePeerMap>,
    self_node: RefCell<Weak<Node>>,
}

impl NodeService {
    pub fn self_node(&self) -> Rc<Node> {
        self.self_node
            .borrow()
            .upgrade()
            .expect("asking for self node before it's been initialized!")
    }

    pub fn build_ephemeral_self_node(&self) {
        assert_eq!(self.self_node.borrow().strong_count(), 0);

        let nself = Rc::new(Node::build_ephemeral());

        self.node_by_pubkey
            .borrow_mut()
            .insert(nself.pubkey.clone(), nself.clone());
        *self.self_node.borrow_mut() = nself.as_weak();
    }

    pub fn node_by_pubkey(&self, pubkey: &PublicKey) -> Option<Rc<Node>> {
        self.node_by_pubkey.borrow().get(pubkey).cloned()
    }

    pub fn peers(&self) -> Vec<Rc<Node>> {
        self.node_by_peer.borrow().values().cloned().collect()
    }

    pub fn inform_of(&self, pubkey: PublicKey, peer: Option<fleck_core::Peer>) {
        // do we know of a node with this pubkey already?
        if let Some(_node) = self.node_by_pubkey(&pubkey) {
            // yes, we do.
            // TODO: update node information if have different peer
        } else {
            // nope! let's add it.
            log::trace!("Informed of new node! pubkey: {:?}", &pubkey);
            let node = Rc::new(Node::build(pubkey.clone(), peer.clone()));

            self.node_by_pubkey
                .borrow_mut()
                .insert(pubkey, node.clone());
            if let Some(peer) = peer {
                self.node_by_peer.borrow_mut().insert(peer, node.clone());
            }
            self.api.channel::<NodeRegistrationChannel>().queue(node);
        }
    }
}

impl Service for NodeService {
    fn new(api: Rc<crate::API>) -> Self {
        Self {
            api,
            node_by_pubkey: Default::default(),
            node_by_peer: Default::default(),
            self_node: Default::default(),
        }
    }

    fn setup(self: &Rc<Self>) {
        self.api.create_channel::<NodeRegistrationChannel>();

        self.api
            .channel::<fleck_core::ReceivePacketChannel>()
            .sub_ref(fleck_core::ReceiveOrder::Annotate, self, Self::annotate);
    }
}

impl NodeService {
    fn annotate(&self, msg: &mut fleck_core::Message) {
        // annotate node based on peer
        if let Some(peer) = &msg.peer {
            let map = self.node_by_peer.borrow();
            match map.get(peer) {
                Some(node) => {
                    msg.node = Some(node.clone());
                },
                None => {
                    // the interesting case, it's a new node!
                    // TODO: send them a request for their public key
                    log::warn!("incoming message from unknown source; peer is {:?}", &peer);
                },
            }
        }
    }
}
