//! `fleck` is a library to for peer-to-peer communications in a distributed system.
//!
//! The core concept is that any program that wants to talk with other instances of itself (or
//! compatible programs) can use `fleck`, and as much of the usual development overhead (network
//! event loop, peer identity verification, packet encryption and forwarding, NAT punchthrough,
//! etc) is handled within the library, in a way that the program need not care about.
//!
//! Currently, the following features are implemented:
//! - Network event loop
//! - Packet encryption
//!
//! ### Library structure
//! `fleck` is structured with a 'global' [`API`] instance, where any node-wide
//! global state is stored. The [`API`] instance stores a collection of
//! [`Channel`](service/event/struct.Channel.html) and [`Service`](service/trait.Service.html)
//! instances. Services listen on specific channels, and events are sent to channels to be handled.
//!
//! For example, the [`NodeService`](fleck_core/node/struct.NodeService.html) will send events on a
//! channel whenever a new node joins the `fleck` network. To reduce shared global state, channels
//! are identified by a static [`ChannelSpec`](service/event/trait.ChannelSpec.html) type, which is
//! usually a tuple of two or three other static types that allow distinguishing between e.g.
//! [messages of different types being received](fleck_core/msg/type.MessageChannel.html), or
//! [coarse](fleck_core/type.MajorTickChannel.html) and
//! [fine](fleck_core/type.MinorTickChannel.html) tick timers.
//!
//! See the documentation for the [`fleck_core`] module to see the core
//! services and channels that make up a standard `fleck` node.

use std::rc::Rc;

pub mod fleck_core;
mod helper;
pub mod service;

pub mod prelude {
    pub use crate::fleck_core;
    pub use crate::helper::{AsAny, IntoWeak};
    pub use crate::service::{ChannelSpec, DefaultService, Service};
}

use prelude::*;

#[cfg(test)]
mod test;

/// Entry point to a running fleck instance.
pub struct API {
    services: std::cell::RefCell<service::ServiceStack>,
}

impl API {
    pub fn new() -> Rc<Self> {
        let res = Rc::new(Self {
            services: Default::default(),
        });

        res.services.borrow_mut().give_api(Rc::downgrade(&res));

        res.add_service::<fleck_core::io::IOService>();
        res.add_service::<fleck_core::CoreInitService>();

        res
    }

    /// Run the fleck instance. Note that this function will not return until fleck is shut down.
    pub fn run(self: &Rc<Self>) {
        self.check_state();

        self.with_service(|io: &fleck_core::io::IOService| {
            io.run();
        });
    }

    /// Run the fleck event-handling loop once and return.
    ///
    /// Note that this will do more work than invoking run(), but may be useful for integration
    /// into other event loops.
    pub fn run_once(self: &Rc<Self>) {
        self.check_state();

        self.with_service(|io: &fleck_core::io::IOService| {
            io.run_once(Some(std::time::Duration::ZERO));
        });
    }

    fn check_state(self: &Rc<Self>) {
        // check that we have a self node
        self.with_service(|node: &fleck_core::NodeService| {
            node.self_node();
        });
    }
}

impl API {
    /// Add a service to the fleck instance.
    pub fn add_service<S: Service + 'static>(&self) {
        self.services.borrow().give_service::<S>();
    }
    /// Access a pre-existing service
    pub fn with_service<S: Service + 'static, R, F: FnOnce(&S) -> R>(&self, f: F) -> R {
        self.services.borrow().with_service(f)
    }
    /// Create an event channel from a specification
    pub fn create_channel<CS: service::ChannelSpec>(
        &self,
    ) -> Rc<service::event::Channel<CS::Tag, CS::Data, CS::Priority>> {
        self.services.borrow().event_root().create_channel::<CS>();
        self.channel::<CS>()
    }
    /// Get an existing channel
    pub fn channel<CS: service::ChannelSpec>(
        &self,
    ) -> Rc<service::event::Channel<CS::Tag, CS::Data, CS::Priority>> {
        self.services.borrow().event_root().channel::<CS>()
    }
    /// Queue a message onto a channel
    pub fn queue<CS: service::ChannelSpec>(&self, data: CS::Data) {
        self.services
            .borrow()
            .event_root()
            .channel::<CS>()
            .as_ref()
            .queue(data);
    }
}
