use std::any::TypeId;
use std::cell::RefCell;
use std::collections::HashMap;
use std::rc::{Rc, Weak};

pub mod event;

pub use event::ChannelSpec;

#[allow(unused_variables)]
pub trait Service: std::any::Any + crate::helper::AsAny {
    fn new(api: Rc<crate::API>) -> Self;
    fn setup(self: &Rc<Self>);

    /// Please don't implement this.
    #[doc(hidden)]
    fn _setup_wrapper(self: &Rc<Self>, _api: Rc<crate::API>) {
        self.setup();
    }
}

#[allow(unused_variables)]
pub trait DefaultService: Default {
    fn setup(self: &Rc<Self>, api: Rc<crate::API>);
}

impl<T: 'static + DefaultService> Service for T {
    fn new(_api: Rc<crate::API>) -> Self {
        Self::default()
    }
    fn setup(self: &Rc<Self>) {}

    fn _setup_wrapper(self: &Rc<Self>, api: Rc<crate::API>) {
        <T as DefaultService>::setup(self, api);
    }
}

#[derive(Default)]
pub struct ServiceStack {
    service_map: RefCell<HashMap<TypeId, Rc<dyn std::any::Any>>>,

    api: Weak<crate::API>,

    eroot: event::EventRoot,
}

impl ServiceStack {
    pub(crate) fn give_api(&mut self, api: Weak<crate::API>) {
        self.api = api;
    }
    pub(crate) fn give_service<S: 'static + Service>(&self) {
        let srv = S::new(self.api.upgrade().expect("no API given"));
        self.add_service(srv);
    }

    pub(crate) fn add_service<S: 'static + Service>(&self, srv: S) {
        let srv = Rc::new(srv);
        srv._setup_wrapper(self.api.upgrade().expect("no API given"));
        self.service_map.borrow_mut().insert(TypeId::of::<S>(), srv);
    }

    pub fn with_service<'l, S: Service + 'static, R, F: 'l + FnOnce(&S) -> R>(&'l self, f: F) -> R {
        let id = TypeId::of::<S>();

        // note: we don't want to hold a reference to service_map when invoking f()
        let svc = {
            let svcs = self.service_map.borrow();
            svcs.get(&id)
                .unwrap_or_else(|| panic!("asked for service {} that doesn't exist!", std::any::type_name::<S>()))
                .clone()
                .downcast::<S>()
                .expect("internal inconsistency: wanted Rc<Service> but that wasn't what was in the map")
        };

        f(svc.as_ref())
    }

    pub fn event_root(&self) -> &event::EventRoot {
        &self.eroot
    }
}
