mod encryption;
mod handshake;
mod routing;

static TRACING_SETUP_FLAG: std::sync::Mutex<bool> = std::sync::Mutex::new(false);

fn setup_tracing() {
    let mut guard = TRACING_SETUP_FLAG.lock().unwrap();
    if !*guard {
        tracing_log::LogTracer::init().unwrap();
        *guard = true;
    }
}

struct TestSkeleton {
    tmpfile: tempfile::NamedTempFile,
}

impl TestSkeleton {
    fn new() -> Self {
        setup_tracing();
        Self {
            tmpfile: tempfile::NamedTempFile::new().expect("couldn't create temporary file?"),
        }
    }

    pub fn socket_path(&self) -> &std::path::Path {
        self.tmpfile.path()
    }

    /// Builds a single UNIX pipe 'server' node and a given number of 'client' nodes
    /// Note that the nodes are initialized and the client nodes are informed of the server's
    /// existence, but they are not stepped forwards at all. If you want the handshake process to
    /// finish, you will have to step them forwards yourself.
    pub fn initialize_nodes(&self, count: usize) -> Vec<std::rc::Rc<crate::API>> {
        assert!(count >= 1);

        let mut result = vec![];

        // initialize server
        result.push({
            use crate::fleck_core;
            let fleck = crate::API::new();

            fleck.with_service(|ns: &fleck_core::NodeService| {
                ns.build_ephemeral_self_node();
            });

            fleck.with_service(|io: &fleck_core::io::IOService| {
                fleck_core::io::UnixSocketBuilder::default()
                    .set_path(self.socket_path().to_str().unwrap())
                    .serve_mode()
                    .build(io);
            });

            fleck
        });

        // initialize clients
        for _ in 1..count {
            result.push({
                use crate::fleck_core;
                let fleck = crate::API::new();

                fleck.with_service(|ns: &fleck_core::NodeService| {
                    ns.build_ephemeral_self_node();
                });

                fleck.with_service(|pd: &fleck_core::discovery::PeerDiscovery| {
                    pd.new_peer(
                        fleck
                            .with_service(|io: &fleck_core::io::IOService| {
                                fleck_core::io::UnixSocketBuilder::default()
                                    .set_path(self.socket_path().to_str().unwrap())
                                    .connect_mode()
                                    .build(io)
                            })
                            .connected_peer(),
                    )
                });

                fleck
            });
        }

        result
    }
}
